Require Import rounding BigZ.

Definition hmult (precision x y : bigZ) := (x * y / precision)%bigZ.
Definition hdiv (precision x y : bigZ) := (x * precision / y)%bigZ.
Definition hsqrt (precision x : bigZ) := BigZ.sqrt (precision * x).
Definition h2 precision := (2 * precision)%bigZ.
Definition h1 (precision : bigZ) := precision.

Fixpoint hpi_rec (precision : bigZ)
  (n : nat) (y z prod : bigZ) {struct n} : bigZ :=
  match n with
  | 0%nat =>
      hmult precision (h2 precision + hsqrt precision (h2 precision)) prod
  | S p =>
      let sy := hsqrt precision y in
      let ny := hdiv precision (h1 precision + y) (2 * sy) in
      let nz :=
        hdiv precision (h1 precision + hmult precision y z)
          (hmult precision (h1 precision + z) sy) in
      hpi_rec precision p ny nz
        (hmult precision prod
           (hdiv precision (h1 precision + ny) (h1 precision + nz)))
  end.

Lemma hpi_rec_step precision n y z prod :
  hpi_rec precision n y z prod =
  match n with
  | 0%nat =>
      hmult precision (h2 precision + hsqrt precision (h2 precision)) prod
  | S p =>
      let sy := hsqrt precision y in
      let ny := hdiv precision (h1 precision + y) (2 * sy) in
      let nz :=
        hdiv precision (h1 precision + hmult precision y z)
          (hmult precision (h1 precision + z) sy) in
      hpi_rec precision p ny nz
        (hmult precision prod
           (hdiv precision (h1 precision + ny) (h1 precision + nz)))
  end.
Proof. destruct n; reflexivity. Qed.

Lemma rounding_hpi_rec_step precision n y z prod :
  rounding.hpi_rec precision n y z prod =
  match n with
  | 0%nat =>
      rounding.hmult precision
        (rounding.h2 precision +
         rounding.hsqrt precision (rounding.h2 precision)) prod
  | S p =>
      let sy := rounding.hsqrt precision y in
      let ny :=
        rounding.hdiv precision (rounding.h1 precision + y) (2 * sy) in
      let nz :=
        rounding.hdiv precision
          (rounding.h1 precision + rounding.hmult precision y z)
          (rounding.hmult precision (rounding.h1 precision + z) sy) in
      rounding.hpi_rec precision p ny nz
        (rounding.hmult precision prod
           (rounding.hdiv precision (rounding.h1 precision + ny)
              (rounding.h1 precision + nz)))
  end.
Proof.
destruct n; reflexivity.
Qed.

Definition hs2 (precision : bigZ) :=
  hsqrt precision (h2 precision).

Definition hy1 (precision : bigZ) := 
  hdiv precision (h1 precision + hs2 precision)
  (2 * hsqrt precision (hs2 precision)).

Definition hz1 (precision : bigZ) :=
  hsqrt precision (hs2 precision).
     
Definition hpi (precision : bigZ) (n : nat) :=
match n with
| 0%nat =>
    (h2 precision + hsqrt precision (h2 precision))%bigZ
| S p =>
    hpi_rec precision p (hy1 precision) (hz1 precision)
      (hdiv precision (h1 precision + hy1 precision)
         (h1 precision + hz1 precision))
end.

Definition million_digit_pi := 
let precision := (2 ^ 3321942)%bigZ in
let n := hpi precision 20 in
let n' := (n * 10 ^ (10 ^ 6 + 4) / 2 ^ 3321942)%bigZ in
let (q, r) := BigZ.div_eucl n' (10 ^ 4) in
(((427 <? r)%bigZ && (r <? 9573)%bigZ)%bool, q).

Require Import Reals.

Definition hR (precision v : bigZ) :=
  rounding.hR [precision]%bigZ [v]%bigZ.

Lemma h2_Z precision : [h2 precision]%bigZ = rounding.h2 [precision]%bigZ.
Proof.
unfold h2, rounding.h2; rewrite BigZ.spec_mul; reflexivity.
Qed.

Lemma hsqrt_Z precision v :
  [hsqrt precision v]%bigZ = rounding.hsqrt [precision]%bigZ [v]%bigZ.
Proof.
unfold hsqrt; rewrite BigZ.spec_sqrt, BigZ.spec_mul; reflexivity.
Qed.

Lemma hmult_Z precision x y :
 [hmult precision x y]%bigZ =
 rounding.hmult [precision]%bigZ [x]%bigZ [y]%bigZ.
Proof.
unfold hmult; rewrite BigZ.spec_div, BigZ.spec_mul; reflexivity.
Qed.

Lemma hdiv_Z precision x y :
 [hdiv precision x y]%bigZ =
 rounding.hdiv [precision]%bigZ [x]%bigZ [y]%bigZ.
Proof.
unfold hdiv; rewrite BigZ.spec_div, BigZ.spec_mul; reflexivity.
Qed.

Lemma hpi_rec_Z precision r y z q :
 [hpi_rec precision r y z q]%bigZ =
  rounding.hpi_rec [precision]%bigZ r [y]%bigZ [z]%bigZ [q]%bigZ.
Proof.
revert y z q; induction r; intros y z q.
 simpl; rewrite hmult_Z, BigZ.spec_add, hsqrt_Z, !h2_Z; reflexivity.
rewrite hpi_rec_step, rounding_hpi_rec_step.
lazy zeta; rewrite IHr.
repeat (rewrite !hdiv_Z || rewrite !BigZ.spec_add || rewrite !BigZ.spec_mul
        || rewrite !hmult_Z || rewrite !hsqrt_Z).
reflexivity.
Qed.

Lemma hz1_Z precision :
  [hz1 precision]%bigZ = rounding.hz1 [precision]%bigZ.
Proof.
unfold hz1, hs2; rewrite !hsqrt_Z, h2_Z; reflexivity.
Qed.

Lemma hy1_Z precision : 
  [hy1 precision]%bigZ = rounding.hy1 [precision]%bigZ.
Proof.
unfold hy1, hs2; rewrite hdiv_Z, BigZ.spec_add, BigZ.spec_mul, !hsqrt_Z, h2_Z.
reflexivity.
Qed.

Lemma hpi_Z precision rank :
  [hpi precision rank]%bigZ = rounding.hpi [precision]%bigZ rank.
Proof.
unfold hpi; destruct rank as [|r].
 simpl; rewrite BigZ.spec_add, hsqrt_Z, !h2_Z; reflexivity.
rewrite hpi_rec_Z, hdiv_Z, !BigZ.spec_add, hz1_Z, hy1_Z.
reflexivity.
Qed.

Lemma million_digit_pi_Z :
  let (b, a) := million_digit_pi in
  rounding.million_digit_pi = (b, [a]%bigZ).
Proof.
 unfold million_digit_pi.
 unfold rounding.million_digit_pi.
 match goal with
  |- context [BigZ.div_eucl ?u ?d] =>
     generalize (BigZ.spec_div_eucl u d);
     case (BigZ.div_eucl u d); intros q r
 end.
assert (tenp4 : [10 ^ 4]%bigZ = (10 ^ 4)%Z).
 rewrite BigZ.spec_pow; reflexivity.
rewrite (BigZ.spec_div (hpi (2 ^ 3321942) 20 * 10 ^ (10 ^ 6 + 4))
            (2 ^ 3321942)).
rewrite (BigZ.spec_mul (hpi (2 ^ 3321942) 20) (10 ^ (10 ^ 6 + 4))).
rewrite (hpi_Z (2 ^ 3321942) 20), (BigZ.spec_pow 2 3321942).
rewrite (BigZ.spec_pow 10 (10 ^ 6 + 4)), (BigZ.spec_add (10 ^ 6) 4).
rewrite (BigZ.spec_pow 10 6), (BigZ.spec_pow 10 4).
assert (q2 : [2]%bigZ = 2%Z) by reflexivity.
assert (q3321942 : [3321942]%bigZ = 3321942%Z) by reflexivity.
assert (q10 : [10]%bigZ = 10%Z) by reflexivity.
assert (q6 : [6]%bigZ = 6%Z) by reflexivity.
assert (q4 : [4]%bigZ = 4%Z) by reflexivity.
rewrite q2, q3321942, q10, q6, q4.
intros t; rewrite <- t, !BigZ.spec_ltb; reflexivity.
Qed.

Lemma pi_osix :
  fst million_digit_pi = true ->
  hR (10 ^ 10 ^ 6) (snd million_digit_pi) < PI <
  hR (10 ^ 10 ^ 6) (snd million_digit_pi) +
  Rpower 10 (- Rpower 10 6).
Proof.
generalize million_digit_pi_Z rounding.pi_osix.
destruct million_digit_pi as [b a]; intros t; rewrite t; intros qb po.
apply qb in po; clear qb; simpl snd in po |- *.
unfold hR; rewrite !BigZ.spec_pow; exact po.
Qed.
